from unittest.mock import patch
import os
import json
from .. import comments_db
from .. import comments

TESTDB = '/tmp/unit_test_comments.db'

# def setup_function():
#     comments_db.init("sqlite:///" + TESTDB)

# def teardown_function():
#     os.remove(TESTDB)

def test_comments_db():
    # comments_db.init("sqlite:///:memory:")
    comments_db.init("sqlite:///" + TESTDB)
    comments_db.set_comments(type=comments_db.C_CONN, dbid='con1', comment='comment of connection')
    comments_db.set_comments(type=comments_db.C_SCHEMA, dbid='con1', schema='schema1', comment='comment of schema')
    comments_db.set_comments(type=comments_db.C_TABLE, dbid='con1', schema='schema1', table='table1', comment='comment of table1')
    comments_db.set_comments(
        type=comments_db.C_COLUMN, dbid='con1', schema='schema1', table='table1', column='c1', comment='comment of c1')
    comments_db.set_comments(
        type=comments_db.C_COLUMN, dbid='con1', schema='schema1', table='table1', column='c2', comment='comment of c2')
    comments_db.set_comments(
        type=comments_db.C_COLUMN, dbid='con1', schema='schema1', table='table1', column='c1', comment='comment of c1-dup')

    c=comments_db.get_conn_comments()
    assert c == {'con1': 'comment of connection'}

    c=comments_db.get_schema_comments('con1')
    assert c == {'schema1': 'comment of schema'}

    c=comments_db.get_table_comments('con1', 'schema1')
    assert c == {'table1': 'comment of table1'}

    c=comments_db.get_column_comments('con1', 'schema1', 'table1')
    assert c == {'c1': 'comment of c1-dup', 'c2': 'comment of c2'}

@patch("jupyterlab_sql_explorer.comments_db.get_column_comments")
def test_comments(mock_get_column_comments):
    comments.init("database::sqlite:///:memory:")
    mock_get_column_comments.return_value={'c1': 'comment c1', 'c2': 'comment c2'}
    columns=[
        {'name': 'c1'},
        {'name': 'c2'},
        {'name': 'c3'},
    ]
    assert comments.match_column('con1', 'schema1', 'table1', columns)==[
        {'name': 'c1', 'desc': 'comment c1'},
        {'name': 'c2', 'desc': 'comment c2'},
        {'name': 'c3'}
    ]

@patch("jupyterlab_sql_explorer.comments_db.get_conn_comments")
def test_comments_1(mock_get_conn_comments):
    comments.init("database::sqlite:///:memory:")
    mock_get_conn_comments.return_value={'con1': 'comment con1'}
    conns=[
        {'name': 'con1'},
        {'name': 'con2'},
    ]
    assert comments.match_conn(conns)==[
        {'name': 'con1', 'desc': 'comment con1'},
        {'name': 'con2'},
    ]

@patch("jupyterlab_sql_explorer.comments_db.get_schema_comments")
def test_comments_2(mock_get_schema_comments):
    comments.init("database::sqlite:///:memory:")
    mock_get_schema_comments.return_value={'s1': 'comment s1'}
    schemas=[
        {'name': 's1'},
        {'name': 's2'},
    ]
    assert comments.match_schema('con1', schemas)==[
        {'name': 's1', 'desc': 'comment s1'},
        {'name': 's2'},
    ]

@patch("jupyterlab_sql_explorer.comments_db.get_table_comments")
def test_comments_3(mock_get_table_comments):
    comments.init("database::sqlite:///:memory:")
    mock_get_table_comments.return_value={'t1': 'comment t1'}
    tables=[
        {'name': 't1'},
        {'name': 't2'},
    ]
    assert comments.match_table('con1', 's1', tables)==[
        {'name': 't1', 'desc': 'comment t1'},
        {'name': 't2'},
    ]

async def test_comments_handler(jp_fetch):
    response = await jp_fetch("jupyterlab-sql-explorer", "comments",
        method='POST', body=json.dumps({"type": 3, "dbid": "con1", "schema": "sch1", "table": "t1", "comment": "-------"}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {
        'data': 'set comment ok'
    }


import json
import os
from unittest.mock import patch
from .. import engine

@patch("jupyterlab_sql_explorer.handlers.engine._getDbInfo")
async def test_sqlite_dbtable(mock_dbinfo, jp_fetch):
    os.system('rm /tmp/jp_sql_tmp.db')
    mock_dbinfo.return_value={'db_id': 'testdb', 'db_type': engine.DB_SQLITE, 'db_name': 'jp_sql_tmp.db'}

    response = await jp_fetch("jupyterlab-sql-explorer", "dbtables", params={'dbid': 'testdb'})
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {"data": []}

    response = await jp_fetch("jupyterlab-sql-explorer", "query",
                              method='POST',
                              body=json.dumps({'dbid': 'testdb', 'sql': 'create table AAA (a int, b int)'}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload['error'] == 'RETRY'

    response = await jp_fetch("jupyterlab-sql-explorer", "query", params={'taskid': payload['data']})
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'data': {}}

    response = await jp_fetch("jupyterlab-sql-explorer", "query",
                              method='POST',
                              body=json.dumps({'dbid': 'testdb', 'sql': 'create view BBB as select * from AAA'}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload['error'] == 'RETRY'

    response = await jp_fetch("jupyterlab-sql-explorer", "query", params={'taskid': payload['data']})
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'data': {}}

    response = await jp_fetch("jupyterlab-sql-explorer", "dbtables", params={'dbid': 'testdb'})
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {"data": [
        {'name': 'AAA', 'desc': '', 'type': 'table', 'subtype': 'T'},
        {'name': 'BBB', 'desc': '', 'type': 'table', 'subtype': 'V'}
    ]}

    response = await jp_fetch("jupyterlab-sql-explorer", "columns", params={'dbid': 'testdb', 'db': '', 'tbl': 'AAA'})
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'data': [{'name': 'a', 'desc': 'INT', 'type': 'col'}, {'name': 'b', 'desc': 'INT', 'type': 'col'}]}


import os
import re
import json
import base64
import sqlalchemy
import gettext
from urllib.parse import quote_plus
_ = gettext.gettext

from .const import DB_ROOT
from . import comments

DB_CFG=DB_ROOT+'db_conf.json'

DB_MYSQL = '1'
DB_PGSQL = '2'
DB_ORACLE = '3'
DB_HIVE_LDAP = '4'
DB_HIVE_KERBEROS = '5'
DB_SQLITE = '6'

_temp_pass_store = dict()

def open_dbfile(dbfile):
    expanded_file_path = os.path.expanduser(dbfile)
    dir_name = os.path.dirname(expanded_file_path)
    os.makedirs(dir_name, exist_ok=True)
    return open(expanded_file_path, 'wt')

def _getDBlist()->list:
    dbs=[]
    for e in os.environ:
        if e[0:3]=='DB_':
            dbs.append(e[3:])
    return dbs

def getDBlist()->list:
    dbs = _getDBlist()
    lst=[]

    for dbid in dbs:
        info = _getDbInfo(dbid)
        subtype = int(info['db_type'])
        lst.append({'name': dbid, 'desc': '', 'type': 'conn', 'subtype': subtype, 'fix': 1})

    for dbid, e in _getCfgEntryList().items():
        lst.append({'name': dbid, 'desc': e['name'], 'type': 'conn', 'subtype': int(e['db_type'])})

    return comments.match_conn(lst)

def _getCfgEntryList(passfile=DB_CFG)->list:
    passfile= os.path.expanduser(passfile)
    if os.path.exists(passfile):
        with open(passfile, mode='rt') as f:
            try:
                dblst = json.load(f)
            except:
                dblist={}
    else:
        dblst={}
    return dblst

def _getCfgEntry(name, passfile=DB_CFG):
    dblst = _getCfgEntryList(passfile)
    if name in dblst:
        return dblst[name]
    return None

def _getDbInfo(name: str)-> 'dict | None':
    var_name='DB_' + name
    db_str=os.getenv(var_name)

    if db_str is not None:
        return json.loads(base64.b64decode(db_str.encode()))

    return _getCfgEntry(name)

def getDbInfo(name: str)-> 'dict | None':
    i = _getDbInfo(name)
    if i is None:
        return None
    if 'db_pass' in i:
        del i['db_pass']
    return i

def _getSQL_engine(dbid, db, usedb=None):

    db_host = db['db_host'] if 'db_host' in db else ''
    db_name = db['db_name'] if 'db_name' in db else ''

    if usedb is not None and db['db_type'] in [DB_MYSQL, DB_HIVE_LDAP, DB_HIVE_KERBEROS]:
        # 对于 myql, hive database / schema 等价
        db_name=usedb

    if db['db_type'] == DB_HIVE_KERBEROS:  # Hive-kerberos
        db_port = db['db_port'] if 'db_port' in db else 10000
        principal = db['principal']
        os.system(f"kinit -kt /opt/conda/etc/keytab_{dbid} {principal}")
        # from pyhive import hive
        # return hive.connect(host=db_host, port=int(db_port), auth='KERBEROS', kerberos_service_name='hive')
        sqlstr = f"hive://{db_host}:{db_port}/{db_name}"
        return sqlalchemy.create_engine(sqlstr, connect_args={'auth': 'KERBEROS', 'kerberos_service_name': 'hive'})

    #
    # set user/pass for db exclude DB_SQLITE
    #
    if db['db_type']!=DB_SQLITE:
        if 'db_user' not in db or 'db_pass' not in db:
            if dbid not in _temp_pass_store:
                db_user = db['db_user'] if 'db_user' in db else None
                input_passwd(dbid, db_user)
                return
            else:
                db_user = _temp_pass_store[dbid]['user']
                db_pass = _temp_pass_store[dbid]['pwd']
        else:
            db_user = db['db_user']
            db_pass = db['db_pass']
        db_pass = quote_plus(db_pass)

    if db['db_type'] == DB_MYSQL:   # MYSQL
        db_port = db['db_port'] if 'db_port' in db else 3306
        sqlstr= f"mysql+pymysql://{db_user}:{db_pass}@{db_host}:{db_port}/{db_name}"

    elif db['db_type'] == DB_PGSQL:  # PGSQL
        db_port = db['db_port'] if 'db_port' in db else 5432
        sqlstr = f"postgresql://{db_user}:{db_pass}@{db_host}:{db_port}/{db_name}"

    elif db['db_type'] == DB_ORACLE:  # ORACLE
        db_port = db['db_port'] if 'db_port' in db else 1521
        sqlstr = f"oracle+cx_Oracle://{db_user}:{db_pass}@{db_host}:{db_port}/{db_name}"

    elif db['db_type'] == DB_HIVE_LDAP:  # Hive-LDAP
        db_port = db['db_port'] if 'db_port' in db else 10000
        # return hive.connect(db_host, port=int(db_port), auth='LDAP', username=db_user, password=db_pass)
        sqlstr = f"hive://{db_user}:{db_pass}@{db_host}:{db_port}/{db_name}"
        return sqlalchemy.create_engine(sqlstr, connect_args={'auth': 'LDAP'})
    elif db['db_type'] == DB_SQLITE:  # SQLITE
        if db_name[0]!='/' and db_name!=':memory:':
            db_name=os.path.expanduser(DB_ROOT+db_name)

        # make sure parent dir exists
        dir_name = os.path.dirname(db_name)
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name, exist_ok=True)

        sqlstr = f"sqlite+pysqlite:///{db_name}"
        return sqlalchemy.create_engine(sqlstr)
    else:
        raise ValueError(("unsupport database type"))

    return sqlalchemy.create_engine(sqlstr, pool_size=20, max_overflow=20, pool_timeout=30000, echo=False)

def __gen_krb5_conf(db):

    if db['db_type'] != DB_HIVE_KERBEROS:
        return

    default_realm = db['def_realm']
    with open('/opt/conda/etc/krb5.conf', 'a') as f:
        f.write(f"[libdefaults]\ndefault_realm = {default_realm}\ndns_lookup_realm = false\ndns_lookup_kdc = false\n\n[realms]\n")
        for realm, cfg in db['krb5conf'].items():
            f.write("  %s = {\n" % realm)
            for k, v in cfg.items():
                f.write(f"    {k} = {v}\n")
            f.write("  }\n")
    # os.system('cat /opt/conda/etc/krb5.conf')

def getEngine(dbid, usedb=None):

    dbinfo = _getDbInfo(dbid)

    if dbinfo is None:
        if os.environ.get('BATCH'):
            print("Can't Access DB: %s" % dbid)
            return False
        newinfo = addEntry(name=dbid)
        return
        __gen_krb5_conf(newinfo)
        return _getSQL_engine(dbid, newinfo, usedb)
    else:
        __gen_krb5_conf(dbinfo)
        return _getSQL_engine(dbid, dbinfo, usedb)

def addEntry(dbinfo, dbfile=DB_CFG):
    dbid = dbinfo['db_id']
    err=''

    if 'db_type' not in dbinfo:
        raise Exception('must set db type.')

    if dbinfo['db_type'] != DB_SQLITE:
        if 'db_host' not in dbinfo:
            err='must set ip addr.'

    if dbinfo['db_type'] == DB_PGSQL:
        if 'db_name' not in dbinfo:
            err='postgres must set database name to connect'

    # check dataname format
    if dbinfo['db_type'] == DB_SQLITE:
        if 'db_name' not in dbinfo:
            err="sqlite must set db name ( it's a database file )"
    elif 'db_name' in dbinfo:
        if not re.match(r'^[a-zA-Z0-9_]+$', dbinfo['db_name']):
            err='db name can only contain letters, numbers, and underscores.'

    fix_dbs =_getDBlist()
    dbcfg=_getCfgEntryList(dbfile)
    if dbid in fix_dbs or dbid in dbcfg:
        err=f'db_id {dbid} already exists.'

    if 'name' not in dbinfo:
        dbinfo['name']=''

    if err!='':
        raise Exception(err)

    dbcfg[dbid]=dbinfo
    cfg=json.dumps(dbcfg, indent=4)
    with open_dbfile(dbfile) as f:
        f.write(cfg)
    return dbinfo

def delEntry(dbid, dbfile=DB_CFG):
    dbcfg=_getCfgEntryList(dbfile)
    if dbid in dbcfg:
        del dbcfg[dbid]
        cfg=json.dumps(dbcfg, indent=4)
        with open_dbfile(dbfile) as f:
            f.write(cfg)

def check_pass(dbid: str)->(bool, str):
    '''
    check passwd is set for dbid
    '''
    dbinfo = _getDbInfo(dbid)
    if dbinfo is None or 'db_type' not in dbinfo:
        raise Exception('conn not exists or error')

    if dbinfo['db_type']==DB_HIVE_KERBEROS or dbinfo['db_type']==DB_SQLITE:
        return (True, None)

    if 'db_user' in dbinfo and 'db_pass' in dbinfo:
        return (True, None)

    if dbid in _temp_pass_store:
        return (True, None)

    db_user = dbinfo['db_user'] if 'db_user' in dbinfo else ''
    return (False, db_user)

def set_pass(dbid: str, user: str, pwd: str)->(bool, str):
    _temp_pass_store[dbid]={'user': user, 'pwd': pwd}

    eng = getEngine(dbid)
    if eng:
        try:
            conn = eng.connect()
            conn.close()
            return True, None
        except:
            del _temp_pass_store[dbid]
            return False, "user or passwd error"
    else:
        del _temp_pass_store[dbid]
        return False, "user or passwd error"

def clear_pass(dbid: str)->None:
    global _temp_pass_store
    if dbid is None or dbid=='':
        _temp_pass_store=dict()
    elif dbid in _temp_pass_store:
        del _temp_pass_store[dbid]



import pytest
from .. import db

def test_limit():
    sql = 'select * from aaa limit 200'
    rc, sql1 = db.set_limit(sql)
    assert sql1=='select * from aaa  LIMIT 200'

    sql = 'select * from aaa limit 200 ORDER by AAA'
    rc, sql1 = db.set_limit(sql)
    assert sql1=='select * from aaa  ORDER by AAA LIMIT 200'

    sql = 'select * from aaa'
    rc, sql1 = db.set_limit(sql)
    assert sql1=='select * from aaa LIMIT 200'

    sql = 'select * from aaa limit 20000'
    rc, sql1 = db.set_limit(sql, 200, 10000)
    assert sql1=='select * from aaa  LIMIT 10000'

    sql = 'select * from aaa limit 1000'
    rc, sql1 = db.set_limit(sql, 200, 10000)
    assert sql1=='select * from aaa  LIMIT 1000'

    sql = 'select * from aaa limit 10'
    rc, sql1 = db.set_limit(sql, 200, 10000)
    assert sql1=='select * from aaa  LIMIT 10'

    sql = '''
    Select * from (
        select * from BBB limit 10
    ) t
    '''
    rc, sql1 = db.set_limit(sql, 200, 10000)
    assert sql1=='''
    Select * from (
        select * from BBB limit 10
    ) t
     LIMIT 200'''

    sql = 'create table aaa (a int, b int)'
    rc, sql1 = db.set_limit(sql, 200, 10000)
    assert sql1=='create table aaa (a int, b int)'

    with pytest.raises(Exception):
        sql = 'select * from AAA limit x 10'
        rc, sql1 = db.set_limit(sql, 200, 10000)


import asyncio
import uuid

task_dict={}

async def create_query_task(*args):
    loop=asyncio.get_event_loop()
    future = loop.run_in_executor(None, *args)
    taskid = str(uuid.uuid4())
    task_dict[taskid] = future
    return taskid

async def get_result(taskid, timeout=118):

    if taskid not in task_dict:
        return False, {'error': 'task not exists'}

    future = task_dict[taskid]
    done, _ = await asyncio.wait({future}, timeout=timeout, return_when=asyncio.FIRST_COMPLETED)
    if future in done:
        result = future.result()
        del task_dict[taskid]
        return True, result
    else:
        return False, {'error': 'RETRY', 'data': taskid}

async def delete(taskid):
    if taskid not in task_dict:
        return False
    future = task_dict[taskid]
    future.cancel()
    del task_dict[taskid]
    return True


import sqlparse
from . import engine
from . import comments
from .serializer import make_row_serializable

log=None
def set_log(_log):
    global log
    log = _log

def query(dbid, sql, **kwargs) ->list:
    '''
    make a query.
    '''
    usedb=None
    if 'db' in kwargs:
        usedb=kwargs['db']
    eng = engine.getEngine(dbid, usedb)
    if eng:
        conn = eng.connect()
        result = conn.exec_driver_sql(sql)
        data = result.fetchall()
        conn.close()
        return data

    return []

def set_limit(sql: str, def_lim: int = 200, max_lim: int = 10000) -> (bool, str):
    '''
    Append LIMIT to a select sql statment.
    If the LIMIT is not set, set the LIMIT to def_lim. If the LIMIT is set and LIMIT < max_limit, keep it unchanged. 
    Otherwise, modify the LIMIT to max_limit.
    '''
    parsed = sqlparse.parse(sql)
    if len(parsed)!=1:
        return False, 'can only process one statement'

    stmt = parsed[0]
    if not isinstance(stmt, sqlparse.sql.Statement) or stmt.get_type() != "SELECT":
        return True, sql

    out=''
    has_limit=False
    after_limit=False
    for token in stmt:
        if has_limit is False:
            if token.ttype == sqlparse.tokens.Keyword and token.value.upper() == "LIMIT":
                has_limit = True  # limit found
            else:
                out += str(token)
        elif after_limit:
            out += str(token)
        else:
            if token.ttype == sqlparse.tokens.Literal.Number.Integer:
                limit_value = int(token.value)
                after_limit = True
            elif token.ttype != sqlparse.tokens.Whitespace:
                raise Exception("Sql Error")

    use_lim = def_lim
    if has_limit:
        if limit_value<=max_lim:
            use_lim = limit_value
        else:
            use_lim = max_lim
    out += f' LIMIT {use_lim}'
    return True, out

def query_exec(dbid, sql, **kwargs) ->dict:
    '''
    make a query, return with header
    '''
    rc, sql = set_limit(sql, 10000)
    if not rc:
        raise Exception(sql)

    usedb=None
    if 'db' in kwargs:
        usedb=kwargs['db']
    eng = engine.getEngine(dbid, usedb)
    if eng:
        conn = eng.connect()
        transaction=conn.begin()
        result = conn.exec_driver_sql(sql)
        transaction.commit()
        conn.close()
        if result.returns_rows:
            data = [make_row_serializable(row) for row in result]
            columns = list(result.keys())
            return {'columns': columns, 'data': data}
    return {}

def get_column_info(dbid, db, tbl):
    '''
    '''
    dbinfo = engine._getDbInfo(dbid)
    if dbinfo is None:
        return

    columns=[]
    eng=engine.getEngine(dbid, db)
    if eng:
        if dbinfo['db_type'] ==engine.DB_SQLITE:
            for r in query(dbid, f"PRAGMA table_info('{tbl}')"):
                columns.append({'name': r[1], 'desc': r[2], 'type': 'col'})
        elif dbinfo['db_type'] ==engine.DB_MYSQL:
            for r in query(dbid, f"SELECT column_name, column_comment FROM information_schema.columns WHERE table_name = '{tbl}' AND table_schema = '{db}'"):
                columns.append({'name': r[0], 'desc': r[1], 'type': 'col'})
        elif dbinfo['db_type'] ==engine.DB_PGSQL:
            for r in query(dbid, '''
                SELECT column_name, data_type, description as comment, table_name
                FROM information_schema.columns
                LEFT JOIN pg_catalog.pg_description 
                    ON (pg_description.objoid = (table_schema || '.' || table_name)::regclass 
                          AND pg_description.objsubid = ordinal_position)
                WHERE table_schema = '%s' and table_name='%s'
                ORDER BY ordinal_position
            ''' %(db, tbl)):
                columns.append({'name': r[0], 'desc': r[2], 'type': 'col'})
        elif dbinfo['db_type'] ==engine.DB_ORACLE:
            for r in query(dbid, f"SELECT column_name, comments FROM all_col_comments WHERE table_name = '${tbl}'"):
                print(r)
                columns.append({'name': r[0], 'desc': '', 'type': 'col'})
        elif dbinfo['db_type'] ==engine.DB_HIVE_LDAP or dbinfo['db_type'] ==engine.DB_HIVE_KERBEROS:
            cols={}
            pk=False
            for r in query(dbid, f"DESCRIBE {tbl}", db=db):
                if r['col_name']=='':
                    continue
                if r['col_name'][0]=='#':
                    if r['col_name']=='# Partition Information':
                        pk=True
                    continue
                if pk is False:
                    cols[r['col_name']]={'name': r['col_name'], 'desc': r['comment'], 'type': 'col'}
                else:
                    cols[r['col_name']]={'name': r['col_name'], 'desc': r['comment'], 'type': 'col', 'stype': 'parkey'}
            columns=list(cols.values())
    columns = comments.match_column(dbid, db, tbl, columns)
    return columns

def get_schema_or_table(dbid, schema):
    '''
    Obtain the schema or table (if there is no scheam layer) of a specified database
    connection
    '''
    dbinfo = engine._getDbInfo(dbid)
    if dbinfo is None:
        return None

    if dbinfo['db_type'] ==engine.DB_SQLITE:
        tables=[]
        for r in query(dbid, '''
            SELECT
                name,
                CASE type
                    WHEN 'view' THEN 'V'
                    ELSE 'T'
                END
            FROM sqlite_master where type='table' or type='view'
        '''):
            tables.append({'name': r[0], 'desc': '', 'type': 'table', 'subtype': r[1]})
        tables = comments.match_table(dbid, '', tables)
        return tables
    elif dbinfo['db_type'] ==engine.DB_PGSQL:
        if schema is None:
            schemas=[]
            for r in query(dbid, "select schema_name from information_schema.schemata where schema_name='public' or schema_owner!='gpadmin'"):
                schemas.append({'name': r[0], 'desc': '', 'type': 'db'})
            schemas = comments.match_schema(dbid, schemas)
            return schemas
        else:
            tables=[]
            for r in query(dbid, '''
            SELECT
                t.table_name,
                CASE t.table_type
                    WHEN 'BASE TABLE' THEN 'T'
                    ELSE 'V'
                END,
                obj_description((t.table_schema || '.' || t.table_name)::regclass, 'pg_class') as comment
            FROM information_schema.tables t
            WHERE t.table_schema='%s'
            ''' % schema):
                tables.append({'name': r[0], 'desc': r[2], 'type': 'table', 'subtype': r[1]})
            tables = comments.match_table(dbid, schema, tables)
            return tables

    elif dbinfo['db_type'] ==engine.DB_MYSQL:
        if schema is None:
            schemas=[]
            for r in query(dbid, "show databases"):
                schemas.append({'name': r[0], 'desc': '', 'type': 'db'})
            schemas = comments.match_schema(dbid, schemas)
            return schemas
        else:
            tables=[]
            for r in query(dbid, '''
                SELECT
                    table_name,
                    table_comment,
                    CASE table_type
                        WHEN 'VIEW' THEN 'V'
                        ELSE 'T'
                    END
                FROM information_schema.tables
                WHERE table_schema = '%s'
            ''' % schema):
                tables.append({'name': r[0], 'desc': r[1], 'type': 'table', 'subtype': r[2]})
            tables = comments.match_table(dbid, schema, tables)
            return tables

    else:
        if schema is None:
            schemas=[]
            for r in query(dbid, "show databases"):
                schemas.append({'name': r[0], 'desc': '', 'type': 'db'})
            schemas = comments.match_schema(dbid, schemas)
            return schemas
        else:
            tables=[]
            for r in query(dbid, "show tables", db=schema):
                tables.append({'name': r[0], 'desc': '', 'type': 'table'})
            tables = comments.match_schema(dbid, schema, tables)
            return tables




import json
import traceback
from jupyter_server.base.handlers import APIHandler
from jupyter_server.utils import url_path_join
import tornado
from . import engine, db, comments
from . import task

class ConnHandler(APIHandler):
    '''
    data source connection handler
    '''
    @tornado.web.authenticated
    def get(self):
        self.finish(json.dumps({'data': engine.getDBlist()}))

    @tornado.web.authenticated
    def post(self):
        try:
            data = self.get_json_body()
            engine.addEntry(data)
            self.finish(json.dumps({'data': engine.getDBlist()}))
        except Exception as err:
            self.log.error(err)
            traceback.print_exc()
            self.finish(json.dumps({'error': str(err)}))

    @tornado.web.authenticated
    def put(self):
        try:
            data = self.get_json_body()
            engine.addEntry(data)
            self.finish(json.dumps({'data': engine.getDBlist()}))
        except Exception as err:
            self.log.error(err)
            traceback.print_exc()
            self.finish(json.dumps({'error': str(err)}))

    @tornado.web.authenticated
    def delete(self):
        dbid=self.get_argument('dbid')
        engine.delEntry(dbid)
        self.finish(json.dumps({'data': engine.getDBlist()}))

class DbTableHandler(APIHandler):
    '''
     Obtain the database or table (if there is no database layer) of a specified database 
     connection
    '''
    @tornado.web.authenticated
    def get(self):
        dbid=self.get_argument('dbid')
        database = self.get_argument('db', None)
        try:
            st, db_user=engine.check_pass(dbid)
            if not st:
                self.finish(json.dumps({'error': 'NEED-PASS', 'pass_info': {'db_id': dbid, 'db_user': db_user}}))
            else:
                data=db.get_schema_or_table(dbid, database)
                self.finish(json.dumps({'data': data}))
        except Exception as err:
            self.log.error(err)
            traceback.print_exc()
            self.finish(json.dumps({'error': "can't get db/table list of "+dbid}))

class TabColumnHandler(APIHandler):
    '''
    Retrieve the schema of a database table.
    '''
    @tornado.web.authenticated
    def get(self):
        dbid = self.get_argument('dbid')
        database = self.get_argument('db', None)
        tbl = self.get_argument('tbl')
        try:
            st, db_user=engine.check_pass(dbid)
            if not st:
                self.finish(json.dumps({'error': 'NEED-PASS', 'pass_info': {'db_id': dbid, 'db_user': db_user}}))
            else:
                data=db.get_column_info(dbid, database, tbl)
                self.finish(json.dumps({'data': data}))
        except Exception as err:
            self.log.error(err)
            traceback.print_exc()
            self.finish(json.dumps({'error': f"can't get table columns of {tbl}, reason: {str(err)}"}))

class PasswdHandler(APIHandler):
    '''
    Retrieve the schema of a database table.
    '''
    @tornado.web.authenticated
    def post(self):
        data = self.get_json_body()
        try:
            st, msg=engine.set_pass(data['db_id'], data['db_user'], data['db_pass'])
            if st:
                self.finish(json.dumps({'data': 'set passwd ok'}))
            else:
                self.finish(json.dumps({'error': msg}))
        except Exception as err:
            self.log.error(err)
            self.finish(json.dumps({'error': "set passwd error : " + data['db_id']}))
    '''
    Clear temporary stored password
    '''
    @tornado.web.authenticated
    def delete(self):
        dbid=self.get_argument('dbid', None)
        engine.clear_pass(dbid)
        self.finish(json.dumps({'data': 'delete pass ok'}))

class QueryHandler(APIHandler):
    '''
    create/get/del query task
    The query may take a long time to run, so we have implemented long polling.
    For queries that have a runtime longer than 120 seconds, the response will be 'RETRY,' and the client will use GET to wait.
    '''
    @tornado.web.authenticated
    async def post(self):
        qdata = self.get_json_body()
        try:
            st, db_user=engine.check_pass(qdata['dbid'])
            if not st:
                self.finish(json.dumps({'error': 'NEED-PASS', 'pass_info': {'db_id': qdata['dbid'], 'db_user': db_user}}))
            else:
                taskid = await task.create_query_task(db.query_exec, qdata['dbid'], qdata['sql'])
                self.finish(json.dumps({'error': 'RETRY', 'data': taskid}))
        except Exception as err:
            self.log.error(err)
            self.finish(json.dumps({'error': str(err)}))

    '''
    get status of query
    '''
    @tornado.web.authenticated
    async def get(self):
        task_id=self.get_argument('taskid')
        try:
            rc, data = await task.get_result(task_id)
            if rc:
                self.finish(json.dumps({'data': data}))
            else:
                self.finish(json.dumps(data))
        except Exception as err:
            self.log.error(err)
            self.finish(json.dumps({'error': str(err)}))
    '''
    canel query
    '''
    @tornado.web.authenticated
    async def delete(self):
        task_id=self.get_argument('taskid')
        try:
            await task.delete(task_id)
            self.finish(json.dumps({}))
        except Exception as err:
            self.log.error(err)
            self.finish(json.dumps({'error': str(err)}))

class CommentsHandler(APIHandler):
    '''
    handler comments
    '''
    @tornado.web.authenticated
    def post(self):
        try:
            data = self.get_json_body()
            msg=comments.add(data)
            self.finish(json.dumps({'data': msg}))
        except Exception as err:
            self.log.error(err)
            traceback.print_exc()
            self.finish(json.dumps({'error': str(err)}))

    @tornado.web.authenticated
    def delete(self):
        try:
            data = self.get_json_body()
            comments.delete(data)
            self.finish(json.dumps({'data': 'set comments ok'}))
        except Exception as err:
            self.log.error(err)
            traceback.print_exc()
            self.finish(json.dumps({'error': str(err)}))

def handler_url(base_url, act):
    return url_path_join(base_url, "jupyterlab-data-explorer", act)

def setup_handlers(web_app):
    host_pattern=".*$"

    base_url=web_app.settings["base_url"]
    handlers=[
        (handler_url(base_url, "conns"), ConnHandler),
        (handler_url(base_url, "dbtables"), DbTableHandler),
        (handler_url(base_url, "columns"), TabColumnHandler),
        (handler_url(base_url, "pass"), PasswdHandler),
        (handler_url(base_url, "query"), QueryHandler),
        (handler_url(base_url, "comments"), CommentsHandler),
    ]
    web_app.add_handlers(host_pattern, handlers)



import datetime
import uuid
import decimal


def make_row_serializable(row):
    return tuple(_make_value_serializable(value) for value in row)


def _make_value_serializable(value):
    type_ = type(value)
    processor = DISPATCHER.get(type_)
    if processor is not None:
        return processor(value)
    else:
        return value


def _uuid_processor(uuid):
    return str(uuid)


def _datetime_processor(dt):
    return str(dt)


def _list_processor(l):
    return [_make_value_serializable(value) for value in l]


def _decimal_processor(dec):
    return str(dec)


DISPATCHER = {
    uuid.UUID: _uuid_processor,
    datetime.datetime: _datetime_processor,
    datetime.date: _datetime_processor,
    list: _list_processor,
    decimal.Decimal: _decimal_processor,
}



from traitlets import Unicode, default
from traitlets.config import Configurable
import os

from ._version import __version__
from .handlers import setup_handlers
from . import db
from . import comments
from .const import DB_ROOT

class JupyterLabSqlExplorer(Configurable):
    """
    Config options for jupyterlab_data_explorer
    """
    comments_store = Unicode(
        help="how to store comments",
        config=True
    )

    @default('comments_store')
    def _comments_store(self):
        p = os.path.expanduser(DB_ROOT+ 'comments.db')
        dir_name = os.path.dirname(p)
        os.makedirs(dir_name, exist_ok=True)
        return "database::sqlite:///" + p

def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": "jupyterlab-data-explorer"
    }]

def _jupyter_server_extension_points():
    return [{
        "module": "jupyterlab_data_explorer"
    }]

def _load_jupyter_server_extension(server_app):
    """Registers the API handler to receive HTTP requests from the frontend extension.

    Parameters
    ----------
    server_app: jupyterlab.labapp.LabApp
        JupyterLab application instance
    """
    cfg = JupyterLabSqlExplorer(config=server_app.config)
    server_app.log.info("use comment store: " + cfg.comments_store)
    comments.init(cfg.comments_store)

    setup_handlers(server_app.web_app)
    name = "jupyterlab_data_explorer"
    server_app.log.info(f"Registered {name} server extension")


# For backward compatibility with notebook server - useful for Binder/JupyterHub
load_jupyter_server_extension = _load_jupyter_server_extension




import os
from . import comments_db
from .const import DB_ROOT

#
# comments_store : None / database / server
#
comments_store = None

def init(store_str):
    global comments_store
    print(store_str)
    arr=store_str.split("::")
    if arr[0]=='database':
        comments_db.init(arr[1])
        comments_store=arr[0]
    else:
        comments_store=None

def add(data):
    if comments_store is None:
        return "can't set comment, please set comment store first!"
    elif comments_store == 'database':
        comments_db.set_comments(**data)
    elif comments_store == 'server':
        pass
    return "set comment ok"

def match_column(dbid: str, schema: str, table: str, columns: list)->list:
    if comments_store is None:
        return columns
    elif comments_store == 'database':
        clist = comments_db.get_column_comments(dbid, schema, table)
    elif comments_store == 'server':
        pass

    for r in columns:
        if r['name'] in clist:
            r['desc']=clist[r['name']]

    return columns

def match_table(dbid: str, schema: str, tables: list)->list:
    if comments_store is None:
        return tables
    elif comments_store == 'database':
        clist = comments_db.get_table_comments(dbid, schema)
    elif comments_store == 'server':
        pass

    for r in tables:
        if r['name'] in clist:
            r['desc']=clist[r['name']]

    return tables

def match_schema(dbid: str, schemas: list)->list:
    if comments_store is None:
        return schemas
    elif comments_store == 'database':
        clist = comments_db.get_schema_comments(dbid)
    elif comments_store == 'server':
        pass

    for r in schemas:
        if r['name'] in clist:
            r['desc']=clist[r['name']]

    return schemas

def match_conn(conns: list)->list:
    if comments_store is None:
        return conns
    elif comments_store == 'database':
        clist = comments_db.get_conn_comments()
    elif comments_store == 'server':
        pass

    for r in conns:
        if r['name'] in clist:
            r['desc']=clist[r['name']]

    return conns


from sqlalchemy import create_engine, Column, Integer, String, Index, text, Sequence
from sqlalchemy.orm import sessionmaker, declarative_base
#from sqlalchemy.ext.declarative import declarative_base

C_CONN = 1
C_SCHEMA = 2
C_TABLE = 3
C_COLUMN = 4

_conn_str=None

Base = declarative_base()

comments_sequence = Sequence('comments_sequence')

class Comments(Base):
    __tablename__ = 'comments'

    id = Column(Integer, comments_sequence, primary_key=True)
    type = Column(Integer)
    dbid = Column(String(20))
    schema = Column(String(100))
    tabname = Column(String(100))
    colname = Column(String(100))
    memo = Column(String(500))

    index_name = Index('idx', type, dbid, schema, tabname, colname)

def init(conn_str: str):
    global _conn_str
    _conn_str= conn_str
    engine = create_engine(conn_str)
    Base.metadata.create_all(engine)

def get_conn_comments():
    try:
        engine = create_engine(_conn_str)
        with engine.connect() as conn:
            result = conn.execute(text('''
            SELECT dbid, memo FROM comments 
            WHERE id in ( SELECT max(id) as id FROM comments
                    WHERE type = :type
                    GROUP BY dbid )
            '''), {"type":str(C_CONN)})
            data={}
            for r in result.fetchall():
                data[r[0]]=r[1]
            return data
    except Exception:
        return {}

def get_schema_comments(dbid: str):
    try:
        engine = create_engine(_conn_str)
        with engine.connect() as conn:
            result = conn.execute(text('''
            SELECT schema, memo FROM comments
            WHERE id in ( SELECT max(id) as id FROM comments
                        WHERE type = :type and dbid = :dbid
                        GROUP BY schema )
            '''), {'type':str(C_SCHEMA), 'dbid':dbid})
            data={}
            for r in result.fetchall():
                data[r[0]]=r[1]
            return data
    except Exception:
        return {}

def get_table_comments(dbid: str, schema: str):
    try:
        engine = create_engine(_conn_str)       
        data={}
        with engine.connect() as conn:
            result = conn.execute(text('''
            SELECT tabname, memo FROM comments
            WHERE id in ( SELECT max(id) as id FROM comments
                        WHERE type = :type and dbid = :dbid and schema = :schema
                        GROUP BY tabname )
            '''), {'type':str(C_TABLE), 'dbid':dbid, 'schema':schema})
            
            for r in result.fetchall():
                data[r[0]]=r[1]
        return data
    except Exception:
        return {}

def get_column_comments(dbid: str, schema: str, table: str):
    try:
        engine = create_engine(_conn_str)
        with engine.connect() as conn:
            result = conn.execute(text('''
            SELECT colname, memo FROM comments
            WHERE id in ( SELECT max(id) as id FROM comments
                        WHERE type = :type and dbid = :dbid and schema = :schema and tabname = :table
                        GROUP BY colname )
            '''), {'type':str(C_COLUMN), 'dbid':dbid, 'schema':schema, 'table':table})
            data={}
            for r in result.fetchall():
                data[r[0]]=r[1]
            return data
    except Exception:
        return {}

def set_comments(**args):

    param={
        'type': args['type'],
        'memo': args['comment'],
        'dbid': args['dbid']
    }

    type = int(args['type'])

    if type==C_CONN:
        pass
    elif type==C_SCHEMA:
        param.update({'schema': args['schema']})
    elif type==C_TABLE:
        param.update({'schema': args['schema'], 'tabname': args['table']})
    elif type==C_COLUMN:
        param.update({'schema': args['schema'], 'tabname': args['table'], 'colname': args['column']})
    else:
        raise Exception('arg error')

    engine = create_engine(_conn_str)
    Session = sessionmaker(bind=engine)
    session = Session()
    c = Comments(**param)
    session.add(c)
    session.commit()


import json
from unittest.mock import patch
from .. import engine

async def test_conn(jp_fetch):
    '''
    test for create/del connetion for database
    '''
    response = await jp_fetch("jupyterlab-sql-explorer", "conns")
    assert response.code == 200
    old = json.loads(response.body)

    response = await jp_fetch("jupyterlab-sql-explorer", "conns",
        method='POST', body=json.dumps({"db_id": "add", "db_name":"/tmp/unit_test.db", "db_type":'6'}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {
        "data": old['data'] + [{'name': 'add', 'desc': '', 'type': 'conn', 'subtype': 6}]
    }

    response = await jp_fetch("jupyterlab-sql-explorer", "conns",
        method='POST', body=json.dumps({"db_id": "add2", "db_name":":memory:", "db_type":'6'}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {
        "data": old['data'] + [
            {'name': 'add', 'desc': '', 'type': 'conn', 'subtype': 6},
            {'name': 'add2', 'desc': '', 'type': 'conn', 'subtype': 6}
        ]
    }

    response = await jp_fetch("jupyterlab-sql-explorer", "conns")
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {
        "data": old['data'] + [
            {'name': 'add', 'desc': '', 'type': 'conn', 'subtype': 6},
            {'name': 'add2', 'desc': '', 'type': 'conn', 'subtype': 6}
        ]
    }

    response = await jp_fetch("jupyterlab-sql-explorer", "dbtables", params={'dbid': 'add'})
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'data': []}

    response = await jp_fetch("jupyterlab-sql-explorer", "conns", method='DELETE', params={'dbid': 'add'})
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {"data": old['data'] + [{'name': 'add2', 'desc': '', 'type': 'conn', 'subtype': 6}]}

async def test_err_conn(jp_fetch):
    '''
    test for create/del connetion for database
    '''
    response = await jp_fetch("jupyterlab-sql-explorer", "conns",
        method='POST', body=json.dumps({"db_id": "add"}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'error': 'must set db type.'}

    response = await jp_fetch("jupyterlab-sql-explorer", "conns",
        method='POST', body=json.dumps({"db_id": "add", "db_name":"/tmp/test.db", "db_type": engine.DB_MYSQL}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'error': 'db name can only contain letters, numbers, and underscores.'}

    response = await jp_fetch("jupyterlab-sql-explorer", "conns",
        method='POST', body=json.dumps({"db_id": "add", "db_host":"192.168.1.100", "db_type": engine.DB_MYSQL}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert 'error' not in payload

    response = await jp_fetch("jupyterlab-sql-explorer", "conns",
        method='POST', body=json.dumps({"db_id": "add", "db_host":"192.168.1.100", "db_type": engine.DB_MYSQL}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'error': 'db_id add already exists.'}


    response = await jp_fetch("jupyterlab-sql-explorer", "conns",
        method='POST', body=json.dumps({"db_id": "add1", "db_type": engine.DB_PGSQL}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'error': 'postgres must set database name to connect'}

    response = await jp_fetch("jupyterlab-sql-explorer", "conns",
        method='POST', body=json.dumps({"db_id": "add1", "db_name": "test", "db_type": engine.DB_PGSQL}))
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'error': 'must set ip addr.'}


@patch("jupyterlab_sql_explorer.handlers.engine._getDbInfo")
async def test_passwd(mock_engine, jp_fetch):
    mock_engine.return_value={'db_id': 'needpass', 'db_user': 'testuser'}
    response = await jp_fetch("jupyterlab-sql-explorer", "columns", params={'dbid': 'needpass', 'db': 'mysql', 'tbl': 'columns_priv'})
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'error': "can't get table columns of columns_priv, reason: conn not exists or error"}

    mock_engine.return_value={'db_id': 'needpass', 'db_type': engine.DB_MYSQL, 'db_user': 'testuser'}
    response = await jp_fetch("jupyterlab-sql-explorer", "columns", params={'dbid': 'needpass', 'db': 'mysql', 'tbl': 'columns_priv'})
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'error': 'NEED-PASS', 'pass_info': {'db_id': 'needpass', 'db_user': 'testuser'}}

    response = await jp_fetch("jupyterlab-sql-explorer", "pass", method='DELETE')
    assert response.code == 200
    payload = json.loads(response.body)
    assert payload == {'data': 'delete pass ok'}